class PullRequestBuildDiffer:
    def __init__(self, pr_num, repo):
        self.pr_num = pr_num
        self.repo = repo
        self._check_environment()

    def diff(self):
        diff_str = self._get_diff_str()
        diffs = get_diffs_by_diff_str(self.repo, diff_str)
        return diffs

    def _check_environment(self):
        raise NotImplementedError

    def _get_diff_str(self):
        raise NotImplementedError


class LocalPullRequestBuildDiffer(PullRequestBuildDiffer):
    # TODO

    def __init__(self):
        raise NotImplementedError


def get_diffs_by_revision(repo, from_revision, to_revision):
    '''Get file changes between two revisions.

    For details on specifying revisions, see `git help revisions`.

    Args:
        repo (git.Repo): Repo object initialized with project root
        from_revision (str): revision identifier for the starting point of the
            diff
        to_revision (str): revision identifier for the ending point of the diff

    Returns:
        list of git.diff.Diff identifying changes between revisions
    '''
    diff_str = '{from_revision}..{to_revision}'.format(
        from_revision=from_revision, to_revision=to_revision)
    return get_diffs_by_diff_str(repo, diff_str)


def get_diff_str_from_commits(a, b):
    return '{a}..{b}'.format(a=a.hexsha, b=b.hexsha)


def get_diffs_by_diff_str(repo, diff_str):
    '''Get file changes via a diff string.

    For details on specifying revisions, see `git help revisions`.

    Args:
        repo (git.Repo): Repo object initialized with project root
        diff_str (str): diff string identifying range of diff. For example,
            `master..HEAD` diffs from master to HEAD, and `12345678..abcdef90`
            compares to commits.

    Returns:
        list of git.diff.Diff identifying changes between revisions
    '''
    a, b = diff_str.split('..')
    a_obj = repo.rev_parse(a)
    b_obj = repo.rev_parse(b)
    diffs = a_obj.diff(b_obj)
    return diffs


# deprecated for now
class PullRequestInfo:
    def __init__(self, pr_num):
        self.pr_num = pr_num

    def _format(self, str):
        return str.format(pr_num=self.pr_num)

    @property
    def local_ref_name(self):
        '''Shorthand name of local ref, e.g. 'pull/1' '''
        return self._format('pull/{pr_num}')

    @property
    def local_rev_name(self):
        '''Full name of revision, e.g. 'refs/heads/pull/1' '''
        return self._format('refs/heads/pull/{pr_num}')

    @property
    def remote_ref_name(self):
        '''Full name of remote ref (as on GitHub), e.g. 'refs/pull/1/head' '''
        return self._format('refs/pull/{pr_num}/head')


class HeadInfo:
    def __init__(self, repo):
        self.head = repo.head

    @property
    def path(self):
        return self.head.ref.path



import copy
import pathlib

import funcy

from ballet.contrib import get_contrib_features
from ballet.exc import UnexpectedValidationStateError
from ballet.feature import Feature
from ballet.util import assertion_method
from ballet.util.gitutil import LocalPullRequestBuildDiffer
from ballet.util.log import logger
from ballet.util.modutil import import_module_at_path, relpath_to_modname
from ballet.util.travisutil import (
    TravisPullRequestBuildDiffer, can_use_travis_differ)

__all__ = [
    'FeatureValidator',
    'PullRequestFeatureValidator'
]


class FeatureValidator:

    def __init__(self, X, y):
        self.X = X
        self.y = y

    def validate(self, feature):
        '''Validate the feature'''
        failures = []
        result = True
        for check, name in self._get_all_checks():
            success = check(feature)
            if not success:
                result = False
                failures.append(name)

        return result, failures

    @assertion_method
    def _is_feature(self, feature):
        assert isinstance(feature, Feature)

    @assertion_method
    def _has_correct_input_type(self, feature):
        '''Check that `input` is a string or iterable of string'''
        input = feature.input
        is_str = funcy.isa(str)
        is_nested_str = funcy.all_fn(
            funcy.iterable, lambda x: all(map(is_str, x)))
        assert is_str(input) or is_nested_str(input)

    @assertion_method
    def _has_transformer_interface(self, feature):
        assert hasattr(feature.transformer, 'fit')
        assert hasattr(feature.transformer, 'transform')

    @assertion_method
    def _can_make_mapper(self, feature):
        try:
            feature.as_dataframe_mapper()
        except Exception:
            raise AssertionError

    @assertion_method
    def _can_fit(self, feature):
        try:
            mapper = feature.as_dataframe_mapper()
            mapper.fit(self.X, self.y)
        except Exception:
            raise AssertionError

    @assertion_method
    def _can_transform(self, feature):
        try:
            mapper = feature.as_dataframe_mapper()
            mapper.fit(self.X, self.y)
            mapper.transform(self.X)
        except Exception:
            raise AssertionError

    @assertion_method
    def _can_fit_transform(self, feature):
        try:
            mapper = feature.as_dataframe_mapper()
            mapper.fit_transform(self.X, self.y)
        except Exception:
            raise AssertionError

    @assertion_method
    def _has_correct_output_dimensions(self, feature):
        try:
            mapper = feature.as_dataframe_mapper()
            X = mapper.fit_transform(self.X, self.y)
        except Exception:
            raise AssertionError

        assert self.X.shape[0] == X.shape[0]

    @assertion_method
    def _can_deepcopy(self, feature):
        try:
            copy.deepcopy(feature)
        except Exception:
            raise AssertionError

    def _get_all_checks(self):
        for method_name in self.__dir__():
            method = getattr(self, method_name)
            if hasattr(method, 'is_check') and method.is_check:
                name = method.__name__
                if name.startswith('_'):
                    name = name[1:]
                yield (method, name)


class PullRequestFeatureValidator:
    APPROPRIATE_CHANGE_TYPES = ['A']
    APPROPRIATE_FILE_EXTS = ['.py']

    def __init__(self, repo, pr_num, contrib_module_path, X, y):
        '''Validate the features introduced in a proposed pull request.

        Args:
            repo (git.Repo): project repo
            pr_num (str): Pull request number
            contrib_module_path (str): Relative path to contrib module
            X (array-like): Example X array-like
            y (array-like): Example y array-like
        '''
        self.repo = repo
        self.pr_num = pr_num
        self.contrib_module_path = contrib_module_path
        self.X = X
        self.y = y

        if can_use_travis_differ():
            self.differ = TravisPullRequestBuildDiffer(self.pr_num)
        else:
            self.differ = LocalPullRequestBuildDiffer()

        # will be set by other methods
        self.file_diffs = None
        self.file_diffs_admissible = None
        self.file_diffs_inadmissible = None
        self.file_diffs_validation_result = None
        self.features = None
        self.features_validation_result = None

    def validate(self):
        '''Validate pull request.

        To do this, follows these steps:
        1. Collects the files that have changed in this pull request as
           compared to a comparison branch.
        2. Categorize these file changes into admissible or inadmissible file
           changes. Admissible file changes solely contribute python files to
           the contrib subdirectory.
        3. Collect features from admissible new files.
        4. Validate each of these features using the FeatureValidator.
        5. Report the overall validation results.
        '''

        # collect, categorize, and validate file changes
        self._collect_file_diffs()
        self._categorize_file_diffs()
        self._validate_files()

        # collect and validate new features
        self._collect_features()
        self._validate_features()

        # determine overall result
        overall_result = self._determine_validation_result()
        return overall_result

    def _collect_file_diffs(self):
        logger.info('Collecting file changes...')

        self.file_diffs = self.differ.diff()

        # log results
        for i, file in enumerate(self.file_diffs):
            logger.debug('File {i}: {file}'.format(i=i, file=file))
        logger.info('Collected {} file(s)'.format(len(self.file_diffs)))

    def _categorize_file_diffs(self):
        '''Partition file changes into admissible and inadmissible changes'''
        if self.file_diffs is None:
            raise UnexpectedValidationStateError(
                'File changes have not been collected.')

        logger.info('Categorizing file changes...')

        self.file_diffs_admissible = []
        self.file_diffs_inadmissible = []

        def is_appropriate_change_type(diff):
            '''File change is an addition'''
            return diff.change_type in \
                PullRequestFeatureValidator.APPROPRIATE_CHANGE_TYPES

        def within_contrib_subdirectory(diff):
            '''File addition is a subdirectory of project's contrib dir'''
            path = diff.b_path
            contrib_relpath = self.contrib_module_path
            try:
                return pathlib.Path(contrib_relpath) in \
                    pathlib.Path(path).parents
            except Exception:
                return False

        def is_appropriate_file_ext(diff):
            '''File change is a python file'''
            path = diff.b_path
            try:
                for ext in PullRequestFeatureValidator.APPROPRIATE_FILE_EXTS:
                    if path.endswith(ext):
                        return True
                return False
            except Exception:
                return False

        is_admissible = funcy.all_fn(
            is_appropriate_change_type,
            within_contrib_subdirectory,
            is_appropriate_file_ext,
        )

        for diff in self.file_diffs:
            if is_admissible(diff):
                self.file_diffs_admissible.append(diff)
                logger.debug(
                    'Categorized {file} as ADMISSIBLE'
                    .format(file=diff.b_path))
            else:
                self.file_diffs_inadmissible.append(diff)
                logger.debug(
                    'Categorized {file} as INADMISSIBLE'
                    .format(file=diff.b_path))

        logger.info('Admitted {} file(s) and rejected {} file(s)'.format(
            len(self.file_diffs_admissible),
            len(self.file_diffs_inadmissible)))

    def _validate_files(self):
        if self.file_diffs_inadmissible is None:
            raise UnexpectedValidationStateError(
                'File diffs have not been categorized.')

        result = len(self.file_diffs_inadmissible) == 0
        self.file_diffs_validation_result = result

    def _collect_features(self):
        if self.file_diffs_admissible is None:
            raise UnexpectedValidationStateError(
                'File diffs have not been collected.')

        logger.info('Collecting features...')

        self.features = []
        for diff in self.file_diffs_admissible:
            path = diff.b_path
            try:
                project_root = pathlib.Path(self.repo.working_tree_dir)
                modname = relpath_to_modname(path)
                modpath = str(project_root.joinpath(path))
                mod = import_module_at_path(modname, modpath)
            except ImportError:
                logger.info(
                    'Validation failure: failed to import module at {}'
                    .format(path))
                logger.exception('Exception details: ')
                self.features_validation_result = False
                continue

            features = get_contrib_features(mod)
            self.features.extend(features)

        logger.info('Collected {} feature(s)'.format(len(self.features)))

    def _validate_features(self):
        if self.features is None:
            raise UnexpectedValidationStateError(
                'Features have not been collected.')

        # if no features were added at all, reject
        if self.features is not None and len(self.features) == 0:
            logger.info('Failed to collect any new feature(s).')
            self.features_validation_result = False
            return

        # get small subset?
        X, y = subsample_data_for_validation(self.X, self.y)

        # validate
        feature_validator = FeatureValidator(X, y)
        overall_result = True
        for feature in self.features:
            result, failures = feature_validator.validate(feature)
            if result is True:
                logger.info(
                    'Feature is valid: {feature}'.format(feature=feature))
            else:
                logger.info(
                    'Feature is NOT valid: {feature}'.format(feature=feature))
                logger.debug(
                    'Failures in validation: {failures}'
                    .format(failures=failures))
                overall_result = False

        self.features_validation_result = overall_result

    def _determine_validation_result(self):
        if self.file_diffs_validation_result is None:
            raise UnexpectedValidationStateError(
                'File diffs have not been validated.')
        if self.features_validation_result is None:
            raise UnexpectedValidationStateError(
                'Feature changes have not been validated.')
        return (self.file_diffs_validation_result and
                self.features_validation_result)


def subsample_data_for_validation(X, y):
    # TODO
    return X, y



import os
from abc import ABCMeta, abstractmethod

import btb
import btb.tuning.gp
import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.externals import joblib
from sklearn.model_selection import (
    KFold, StratifiedKFold, cross_val_score, cross_validate, train_test_split)
from sklearn.model_selection._validation import _multimetric_score
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from ballet.modeling.constants import RANDOM_STATE
from ballet.modeling.io_transformers import (
    FeatureTypeTransformer, TargetTypeTransformer)
from ballet.modeling.scoring import (
    ScorerInfo, get_scorer_names_for_problem_type)
from ballet.util.log import logger


class Modeler:
    '''Versatile modeling object.

    Handles classification and regression problems and computes variety of
    performance metrics.

    Args:
        problem_type (ProblemType)
    '''

    def __init__(self, problem_type=None, scorers=None):
        self.problem_type = problem_type

        if scorers is None:
            names = get_scorer_names_for_problem_type(self.problem_type)
        self.scorers_info = [
            ScorerInfo(name=name)
            for name in names
        ]
        self.scorers = [
            scorer_info.scorer
            for scorer_info in self.scorers_info
        ]
        self.primary_scorer = self.scorers[0]

        self.estimator = self._get_default_estimator()
        self.feature_type_transformer = FeatureTypeTransformer()

        needs_label_binarizer = self.problem_type.multi_classification
        self.target_type_transformer = TargetTypeTransformer(
            needs_label_binarizer=needs_label_binarizer)

    def set_estimator(self, estimator):
        self.estimator = estimator

    def fit(self, X, y, **kwargs):
        X, y = self._format_inputs(X, y)
        self.estimator.fit(X, y, **kwargs)

    def predict(self, X):
        X = self._format_X(X)
        return self.estimator.predict(X)

    def predict_proba(self, X):
        X = self._format_X(X)
        return self.estimator.predict_proba(X)

    def score(self, X, y):
        X, y = self._format_inputs(X, y)
        return self.estimator.score(X, y)

    def dump(self, filepath):
        joblib.dump(self.estimator, filepath, compress=True)

    def load(self, filepath):
        if not os.path.exists(filepath):
            raise ValueError("Couldn't find model at {}".format(filepath))
        self.estimator = joblib.load(filepath)

    def compute_metrics_cv(self, X, y, **kwargs):
        '''Compute cross-validated metrics.

        Trains this model on data X with labels y.
        Returns a list of dict with keys name, scoring_name, value.

        Args:
        X (Union[np.array, pd.DataFrame]): data
        y (Union[np.array, pd.DataFrame, pd.Series]): labels
        '''

        # compute scores
        results = self.cv_score_mean(X, y)
        return results

    def _compute_metrics_train_test(self, X_tr, y_tr, X_te, y_te):
        '''Compute metrics on test set, given entire train-test split'''
        X_tr, y_tr = self._format_inputs(X_tr, y_tr)
        X_te, y_te = self._format_inputs(X_te, y_te)

        # fit model on entire training set
        self.estimator.fit(X_tr, y_tr)

        scorers = {
            scorer_info.name: scorer_info.scorer
            for scorer_info in self.scorers_info
        }
        multimetric_score_results = _multimetric_score(
            self.estimator, X_te, y_te, scorers)

        return self._process_cv_results(
            multimetric_score_results, filter_testing_keys=False)

    def compute_metrics_train_test(self, X, y, n):
        '''Compute metrics on test set, doing train-test split on inputs'''
        X_tr, X_te, y_tr, y_te = train_test_split(
            X, y, train_size=n, test_size=len(y) - n, shuffle=True)
        return self._compute_metrics_train_test(X_tr, y_tr, X_te, y_te)

    def cv_score_mean(self, X, y):
        '''Compute mean score across cross validation folds.

        Split data and labels into cross validation folds and fit the model for
        each fold. Then, for each scoring type in scorings, compute the score.
        Finally, average the scores across folds. Returns a dictionary mapping
        scoring to score.

        Args:
        X (np.array): data
        y (np.array): labels
        scorings (List[str]): scoring types
        '''

        X, y = self._format_inputs(X, y)

        if self.problem_type.binary_classification:
            kf = StratifiedKFold(
                shuffle=True, random_state=RANDOM_STATE + 3)
        elif self.problem_type.multi_classification:
            self.target_type_transformer.inverse_transform(y)
            transformer = self.target_type_transformer
            kf = StratifiedKFoldMultiClassIndicator(
                transformer, shuffle=True, n_splits=3,
                random_state=RANDOM_STATE + 3)
        elif self.problem_type.regression:
            kf = KFold(shuffle=True, n_splits=3, random_state=RANDOM_STATE + 4)
        else:
            raise NotImplementedError

        scoring = {
            scorer_info.name: scorer_info.scorer
            for scorer_info in self.scorers_info
        }
        cv_results = cross_validate(
            self.estimator, X, y,
            scoring=scoring, cv=kf, return_train_score=False)

        # post-processing
        results = self._process_cv_results(cv_results)
        return results

    def _process_cv_results(self, cv_results, filter_testing_keys=True):
        result = []
        for key, val in cv_results.items():
            if filter_testing_keys:
                if key.startswith('test_'):
                    scorer_name = key[len('test_'):]
                else:
                    continue
            else:
                scorer_name = key
            scorer_description = ScorerInfo(name=scorer_name).description
            val = np.nanmean(cv_results[key])
            if np.isnan(val):
                val = None
            result.append({
                'name': scorer_name,
                'description': scorer_description,
                'value': val,
            })

        return result

    def _format_inputs(self, X, y):
        return self._format_X(X), self._format_y(y)

    def _format_y(self, y):
        return self.target_type_transformer.fit_transform(y)

    def _format_X(self, X):
        return self.feature_type_transformer.fit_transform(X)

    def _get_default_estimator(self):
        if self.problem_type.classification:
            return self._get_default_classifier()
        elif self.problem_type.regression:
            return self._get_default_regressor()
        else:
            raise NotImplementedError

    def _get_default_classifier(self):
        return RandomForestClassifier(
            n_estimators=10, random_state=RANDOM_STATE + 1)

    def _get_default_regressor(self):
        return RandomForestRegressor(
            n_estimators=10, random_state=RANDOM_STATE + 2)


class DecisionTreeModeler(Modeler):

    def _get_default_classifier(self):
        return DecisionTreeClassifier(random_state=RANDOM_STATE + 1)

    def _get_default_regressor(self):
        return DecisionTreeRegressor(random_state=RANDOM_STATE + 2)


class SelfTuningMixin(metaclass=ABCMeta):

    @abstractmethod
    def get_tunables(self):
        return None

    @property
    def tunables(self):
        if not hasattr(self, '_tunables'):
            self._tunables = self.get_tunables()
        return self._tunables

    @tunables.setter
    def tunables(self, tunables):
        self._tunables = tunables

    @property
    def tuning_cv(self):
        if not hasattr(self, '_tuning_cv'):
            self._tuning_cv = 3
        return self._tuning_cv

    @tuning_cv.setter
    def tuning_cv(self, tuning_cv):
        self._tuning_cv = tuning_cv

    @property
    def tuning_iter(self):
        if not hasattr(self, '_tuning_iter'):
            self._tuning_iter = 3
        return self._tuning_iter

    @tuning_iter.setter
    def tuning_iter(self, tuning_iter):
        self._tuning_iter = tuning_iter

    def _get_parent_instance(self):
        # this is probably a sign of bad design pattern
        mro = type(self).mro()
        ParentClass = mro[mro.index(__class__) + 1]  # noqa
        return ParentClass()

    def fit(self, X, y, tune=True, **fit_kwargs):
        if tune:
            # do some tuning
            if self.tunables is not None:

                scorer = None

                def score(estimator):
                    scores = cross_val_score(
                        estimator, X, y,
                        scoring=scorer, cv=self.tuning_cv,
                        fit_params=fit_kwargs)
                    return np.mean(scores)

                logger.info('Tuning model using BTB GP tuner...')
                tuner = btb.tuning.gp.GP(self.tunables)
                estimator = self._get_parent_instance()
                original_score = score(estimator)
                # TODO: this leads to an error because default value of
                # max_depth for RF is `None`
                # params = funcy.project(
                #     estimator.get_params(), [t[0] for t in self.tunables])
                # tuner.add(params, original_score)
                for i in range(self.tuning_iter):
                    params = tuner.propose()
                    estimator.set_params(**params)
                    score_ = score(estimator)
                    logger.debug(
                        'Iteration {}, params {}, score {}'
                        .format(i, params, score_)
                    )
                    tuner.add(params, score_)

                best_params = tuner._best_hyperparams
                best_score = tuner._best_score
                self.set_params(**best_params)
                logger.info(
                    'Tuning complete. '
                    'Cross val score changed from {:0.3f} to {:0.3f}.'
                    .format(original_score, best_score))
            else:
                logger.warning('Tuning requested, but either btb not '
                               'installed or tunable HyperParameters not '
                               'specified.')

        return super().fit(X, y, **fit_kwargs)


class SelfTuningRandomForestMixin(SelfTuningMixin):

    def get_tunables(self):
        return [
            ('n_estimators',
             btb.HyperParameter(btb.ParamTypes.INT, [10, 500])),
            ('max_depth',
             btb.HyperParameter(btb.ParamTypes.INT, [3, 20]))
        ]


class TunedRandomForestRegressor(
        SelfTuningRandomForestMixin, RandomForestRegressor):
    pass


class TunedRandomForestClassifier(
        SelfTuningRandomForestMixin, RandomForestClassifier):
    pass


class TunedModeler(Modeler):

    def _get_default_classifier(self):
        return TunedRandomForestClassifier(random_state=RANDOM_STATE + 1)

    def _get_default_regressor(self):
        return TunedRandomForestRegressor(random_state=RANDOM_STATE + 2)


class StratifiedKFoldMultiClassIndicator(StratifiedKFold):
    '''Adaptation of StratifiedKFold for multiclass-indicator format y values.

    Note that this should not be used for multilabel, multiclass data.
    '''

    def __init__(self, transformer, *args, **kwargs):
        self.transformer = transformer
        super().__init__(*args, **kwargs)

    def split(self, X, y, groups=None):
        y = self.transformer.inverse_transform(y)
        return super().split(X, y, groups=None)



class LoggingContext(object):
    '''
    Logging context manager

    Source: <https://docs.python.org/3/howto/logging-cookbook.html #using-a-context-manager-for-selective-logging>  # noqa
    '''

    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler is not None:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler is not None:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions



from sklearn.pipeline import FeatureUnion

from ballet.eng.base import GroupedFunctionTransformer

__all__ = ['SingleLagger', 'make_multi_lagger']


class SingleLagger(GroupedFunctionTransformer):
    def __init__(self, lag, groupby_kwargs=None):
        super().__init__(lambda x: x.shift(lag), groupby_kwargs=groupby_kwargs)


def make_multi_lagger(lags, groupby_kwargs=None):
    laggers = [SingleLagger(l, groupby_kwargs=groupby_kwargs) for l in lags]
    feature_union = FeatureUnion([
        (repr(lagger), lagger) for lagger in laggers
    ])
    return feature_union



import funcy
import numpy as np

RANDOM_STATE = 1754


def asarray2d(a):
    arr = np.asarray(a)
    if arr.ndim == 1:
        arr = arr.reshape(-1, 1)
    return arr


def get_arr_desc(arr):
    desc = '{typ} {shp}'
    typ = type(arr)
    shp = getattr(arr, 'shape', None)
    return desc.format(typ=typ, shp=shp)


def indent(text, n=4):
    _indent = ' ' * n
    return '\n'.join([_indent + line for line in text.split('\n')])


def assertion_method(func):
    '''Evaluate func, returning T if no errors and F if AssertionError'''
    @funcy.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except AssertionError:
            return False
    wrapped.is_check = True
    return wrapped



import os

import funcy
import git

from ballet.exc import UnexpectedTravisEnvironmentError
from ballet.util.gitutil import PullRequestBuildDiffer
from ballet.util.log import logger


def get_travis_env_or_fail(name):
    if name in os.environ:
        return os.environ[name]
    else:
        # dump_travis_env_vars()
        raise UnexpectedTravisEnvironmentError


def ensure_expected_travis_env_vars(names):
    if not set(names).issubset(set(os.environ.keys())):
        raise UnexpectedTravisEnvironmentError


def dump_travis_env_vars():
    travis_env_vars = funcy.select_keys(
        lambda key: key.startswith('TRAVIS'),
        os.environ)
    logger.debug(travis_env_vars)


def get_travis_pr_num():
    '''Return the PR number if the job is a pull request, None otherwise

    See also:
        - <https://docs.travis-ci.com/user/environment-variables
          /#Default-Environment-Variables>
    '''
    try:
        travis_pull_request = get_travis_env_or_fail('TRAVIS_PULL_REQUEST')
        if travis_pull_request == 'false':
            return None
        else:
            try:
                pr_num = int(travis_pull_request)
                return pr_num
            except ValueError:
                return None
    except UnexpectedTravisEnvironmentError:
        return None


def is_travis_pr():
    '''Check if the current job is a pull request build'''
    return get_travis_pr_num() is not None


def can_use_travis_differ():
    '''Check if the required travis env vars are set for the travis differ'''
    try:
        ensure_expected_travis_env_vars(
            TravisPullRequestBuildDiffer.EXPECTED_TRAVIS_ENV_VARS)
        return True
    except UnexpectedTravisEnvironmentError:
        return False


class TravisPullRequestBuildDiffer(PullRequestBuildDiffer):
    EXPECTED_TRAVIS_ENV_VARS = [
        'TRAVIS_BUILD_DIR',
        'TRAVIS_PULL_REQUEST',
        'TRAVIS_COMMIT_RANGE',
    ]

    def __init__(self, pr_num, repo=None):
        if repo is None:
            repo = self._detect_repo()
        super().__init__(pr_num, repo)

    def _check_environment(self):
        ensure_expected_travis_env_vars(
            TravisPullRequestBuildDiffer.EXPECTED_TRAVIS_ENV_VARS)

        travis_pr_num = get_travis_pr_num()
        if travis_pr_num != self.pr_num:
            raise UnexpectedTravisEnvironmentError

    def _get_diff_str(self):
        return get_travis_env_or_fail('TRAVIS_COMMIT_RANGE')

    def _detect_repo(self):
        build_dir = get_travis_env_or_fail('TRAVIS_BUILD_DIR')
        return git.Repo(build_dir)



import importlib
import pkgutil
import types

import funcy

from ballet.feature import Feature
from ballet.util.log import logger

__all__ = [
    'get_contrib_features',
]


def get_contrib_features(contrib):
    '''Get contributed features from within given module

    Be very careful with untrusted code. The module/package will be
    walked, every submodule will be imported, and all the code therein will be
    executed. But why would you be trying to import from an untrusted package
    anyway?

    Args:
        contrib (module): module (standalone or package) that contains feature
            definitions

    Returns:
        List of Feature
    '''

    if isinstance(contrib, types.ModuleType):
        contrib_features = []

        # fuuuuu
        importlib.invalidate_caches()

        # any module that has a __path__ attribute is a package
        if hasattr(contrib, '__path__'):
            features = get_contrib_features_from_package(contrib)
            contrib_features.extend(features)
        else:
            features = get_contrib_features_from_module(contrib)
            contrib_features.extend(features)
        return contrib_features
    else:
        raise ValueError('Input is not a module')


def get_contrib_features_from_package(package):
    contrib_features = []

    logger.debug(
        'Walking package path {path} to detect modules...'
        .format(path=package.__path__))
    for importer, modname, _ in pkgutil.walk_packages(
            path=package.__path__,
            prefix=package.__name__ + '.',
            onerror=logger.error):
        try:
            mod = importer.find_module(modname).load_module(modname)
        except ImportError:
            logger.exception(
                'Failed to import module {modname}'
                .format(modname=modname))
            continue
        features = get_contrib_features_from_module(mod)
        contrib_features.extend(features)

    return contrib_features


def get_contrib_features_from_module(mod):
    contrib_features = []

    logger.debug(
        'Trying to import contributed feature(s) from module {modname}...'
        .format(modname=mod.__name__))

    # case 1: file defines `features` variable
    try:
        features = import_contrib_feature_from_collection(mod)
        contrib_features.extend(features)
        logger.debug(
            'Imported {n} feature(s) from {modname} from collection'
            .format(n=len(features), modname=mod.__name__))
    except ImportError:
        # case 2: file has at least `input` and `transformer` defined
        try:
            feature = import_contrib_feature_from_components(mod)
            contrib_features.append(feature)
            logger.debug(
                'Imported 1 feature from {modname} from components'
                .format(modname=mod.__name__))
        except ImportError:
            # case 3: nothing useful in file
            logger.debug(
                'Failed to import anything useful from module {modname}'
                .format(modname=mod.__name__))

    return contrib_features


def import_contrib_feature_from_components(mod):
    required = ['input', 'transformer']
    optional = ['name', 'description', 'output', 'options']
    required_vars, optional_vars = import_names_from_module(
        mod, required, optional)
    feature = Feature(
        input=required_vars['input'],
        transformer=required_vars['transformer'],
        source=mod.__name__,
        **optional_vars)
    return feature


def import_contrib_feature_from_collection(mod):
    required = 'features'
    optional = None
    required_vars, _ = import_names_from_module(
        mod, required, optional)
    features = required_vars['features']
    for feature in features:
        feature.source = mod.__name__
    return features


def import_names_from_module(mod, required, optional):

    msg = funcy.partial(
        'Required variable {varname} not found in module {modname}'
        .format, modname=mod.__name__)

    # required vars
    if required:
        required_vars = {}
        if isinstance(required, str):
            required = [required]
        for varname in required:
            if hasattr(mod, varname):
                required_vars[varname] = getattr(mod, varname)
            else:
                raise ImportError(msg(varname=varname))
    else:
        required_vars = None

    # optional vars
    if optional:
        if isinstance(optional, str):
            optional = [optional]
        optional_vars = {k: getattr(mod, k)
                         for k in optional if hasattr(mod, k)}
    else:
        optional_vars = None

    return required_vars, optional_vars



class ProblemType:
    classification = False
    binary_classification = False
    multi_classification = False
    regression = False


class ClassificationProblem(ProblemType):
    classification = True


class BinaryClassificationProblem(ClassificationProblem):
    binary_classification = True


class MulticlassClassificationProblem(ClassificationProblem):
    multi_classification = True


class RegressionProblem(ProblemType):
    regression = True



import unittest


class TestImports(unittest.TestCase):
    def test_ts_imports(self):
        from ballet.eng.ts import (  # noqa
            SingleLagger, make_multi_lagger, )

    def test_missing_imports(self):
        from ballet.eng.missing import (  # noqa
            LagImputer, NullFiller, NullIndicator, )

    def test_base_imports(self):
        from ballet.eng.base import (  # noqa
            NoFitMixin, SimpleFunctionTransformer,
            GroupedFunctionTransformer, )

    def test_misc_imports(self):
        from ballet.eng.misc import (  # noqa
            IdentityTransformer, ValueReplacer, NamedFramer, )

    def test_all_imports(self):
        from ballet.eng import (  # noqa
            SingleLagger, make_multi_lagger,
            LagImputer, NullFiller, NullIndicator,
            NoFitMixin, SimpleFunctionTransformer, GroupedFunctionTransformer,
            IdentityTransformer, ValueReplacer, NamedFramer, )

    def test_nonexistent_import(self):
        with self.assertRaises(ImportError):
            from ballet.eng import bob  # noqa



import sklearn.metrics
from funcy import fallback, flip, partial, raiser, rcompose, select_values

from ballet.exc import Error
from ballet.modeling.constants import (
    CLASSIFICATION_SCORING, REGRESSION_SCORING, SCORING_NAME_MAPPER)
from ballet.util.log import logger
from ballet.util.modutil import import_module_from_modname


class ScorerInfo:

    def __init__(self, scorer=None, name=None, description=None):
        self._scorer = scorer
        self._name = name
        self._description = description

    @property
    def scorer(self):
        def from_attr():
            if self._scorer is not None:
                return self._scorer
            else:
                raise ValueError

        def from_sklearn_metrics():
            return sklearn.metrics.get_scorer(self.name)

        def from_import():
            module_name, scorer_name = self.name.rsplit('.', maxsplit=1)
            mod = import_module_from_modname(module_name)
            return getattr(mod, scorer_name)

        self._scorer = fallback(
            from_attr,
            from_sklearn_metrics,
            from_import,
            raiser(
                Error,
                'Could not get a scorer with configuration {}'.format(
                    self.name)),
        )
        return self._scorer

    @property
    def name(self):
        if self._name is None:
            if self._scorer is not None:
                # try from scorer
                if isinstance(self._scorer,
                              sklearn.metrics.scorer._BaseScorer):
                    scorers = sklearn.metrics.scorer.SCORERS
                    matches = select_values(
                        lambda x: x == self._scorer, scorers)
                    matches = list(matches.keys())
                    if len(matches) == 1:
                        self._name = matches[0]
                    elif len(matches) > 1:
                        # unexpected
                        logger.debug(
                            'Unexpectedly found multiple matches for scorer '
                            'name {name}: {matches!r}'
                            .format(name=self._name, matches=matches))
                    else:
                        # must be a custom scorer, try to get name
                        if hasattr(self._scorer, '__name__'):
                            self._name = self._scorer.__name__
            elif self._description is not None:
                # try from description
                mapper = flip(SCORING_NAME_MAPPER)
                if self._description in mapper:
                    self._name = mapper[self._description]
                else:
                    # default formatting
                    self._name = '_'.join(self._description.lower().split(' '))

        if self._name is not None:
            return self._name
        else:
            raise Error('Could not get name from scorer')

    @property
    def description(self):
        if self._description is None:
            name = self.name

            if name in SCORING_NAME_MAPPER:
                self._description = SCORING_NAME_MAPPER[name]
            else:
                # default formatting
                def upper_first(s):
                    return s[0].upper() + s[1:] if s is not None else s
                format = rcompose(
                    lambda s: s.split('_'),
                    partial(map, upper_first),
                    lambda l: ' '.join(l),
                )
                self._description = format(name)

        return self._description


def get_scorer_names_for_problem_type(problem_type):
    '''Get scorers for this problem type.

    Returns:
        list: List of scorer_name as defined in sklearn.metrics. This is a
            'utility variable' that can be used where we just need the names of
             the scoring functions and not the more complete information.
    '''
    # scoring_types maps user-readable name to `scoring`, as argument to
    # cross_val_score
    # See also
    # http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
    if problem_type.classification:
        return CLASSIFICATION_SCORING
    elif problem_type.regression:
        return REGRESSION_SCORING
    else:
        raise NotImplementedError



import funcy
import numpy as np
import pandas as pd

from ballet.eng.base import BaseTransformer, SimpleFunctionTransformer
from ballet.util import get_arr_desc

__all__ = ['IdentityTransformer', 'ValueReplacer', 'NamedFramer']


class IdentityTransformer(SimpleFunctionTransformer):
    def __init__(self):
        super().__init__(funcy.identity)


class ValueReplacer(BaseTransformer):
    def __init__(self, value, replacement):
        super().__init__()
        self.value = value
        self.replacement = replacement

    def transform(self, X, **transform_kwargs):
        X = X.copy()
        mask = X == self.value
        X[mask] = self.replacement
        return X


class NamedFramer(BaseTransformer):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def transform(self, X, **transform_kwargs):
        msg = "Couldn't convert object {} to named 1d DataFrame."
        if isinstance(X, pd.Index):
            return X.to_series().to_frame(name=self.name)
        elif isinstance(X, pd.Series):
            return X.to_frame(name=self.name)
        elif isinstance(X, pd.DataFrame):
            if X.shape[1] == 1:
                X = X.copy()
                X.columns = [self.name]
                return X
            else:
                raise ValueError(msg.format(get_arr_desc(X)))
        elif isinstance(X, np.ndarray):
            if X.ndim == 1:
                return pd.DataFrame(data=X.reshape(-1, 1), columns=[self.name])
            elif X.ndim == 2 and X.shape[1] == 1:
                return pd.DataFrame(data=X, columns=[self.name])
            else:
                raise ValueError(msg.format(get_arr_desc(X)))

        raise TypeError(msg.format(get_arr_desc(X)))



import unittest
from textwrap import dedent
from unittest.mock import patch

import funcy
import numpy as np
import pandas as pd

from ballet.compat import SimpleImputer
from ballet.exc import UnexpectedValidationStateError
from ballet.feature import Feature
from ballet.eng.base import BaseTransformer
from ballet.eng.misc import IdentityTransformer
from ballet.util.gitutil import get_diff_str_from_commits
from ballet.util.travisutil import TravisPullRequestBuildDiffer
from ballet.validation import FeatureValidator, PullRequestFeatureValidator

from .util import (
    FragileTransformer, make_mock_commit, make_mock_commits, mock_repo)


class TestDataMixin:
    def setUp(self):
        self.df = pd.DataFrame(
            data={
                'country': ['USA', 'USA', 'Canada', 'Japan'],
                'year': [2001, 2002, 2001, 2002],
                'size': [np.nan, -11, 12, 0.0],
                'strength': [18, 110, np.nan, 101],
                'happy': [False, True, False, False]
            }
        ).set_index(['country', 'year'])
        self.X = self.df[['size', 'strength']]
        self.y = self.df[['happy']]
        super().setUp()


class TestFeatureValidator(TestDataMixin, unittest.TestCase):

    def test_good_feature(self):
        feature = Feature(
            input='size',
            transformer=SimpleImputer(),
        )

        validator = FeatureValidator(self.X, self.y)
        result, failures = validator.validate(feature)
        self.assertTrue(result)
        self.assertEqual(len(failures), 0)

    def test_bad_feature_input(self):
        # bad input
        feature = Feature(
            input=3,
            transformer=SimpleImputer(),
        )
        validator = FeatureValidator(self.X, self.y)
        result, failures = validator.validate(feature)
        self.assertFalse(result)
        self.assertIn('has_correct_input_type', failures)

    def test_bad_feature_transform_errors(self):
        # transformer throws errors
        feature = Feature(
            input='size',
            transformer=FragileTransformer(
                (lambda x: True, ), (RuntimeError, ))
        )
        validator = FeatureValidator(self.X, self.y)
        result, failures = validator.validate(feature)
        self.assertFalse(result)
        self.assertIn('can_transform', failures)

    def test_bad_feature_wrong_transform_length(self):
        class _WrongLengthTransformer(BaseTransformer):
            def transform(self, X, **transform_kwargs):
                new_shape = list(X.shape)
                new_shape[0] += 1
                output = np.arange(np.prod(new_shape)).reshape(new_shape)
                return output

        # doesn't return correct length
        feature = Feature(
            input='size',
            transformer=_WrongLengthTransformer(),
        )
        validator = FeatureValidator(self.X, self.y)
        result, failures = validator.validate(feature)
        self.assertFalse(result)
        self.assertIn('has_correct_output_dimensions', failures)

    def test_bad_feature_deepcopy_fails(self):
        class _CopyFailsTransformer(IdentityTransformer):
            def __deepcopy__(self):
                raise RuntimeError
        feature = Feature(
            input='size',
            transformer=_CopyFailsTransformer(),
        )
        validator = FeatureValidator(self.X, self.y)
        result, failures = validator.validate(feature)
        self.assertFalse(result)
        self.assertIn('can_deepcopy', failures)


class TestPullRequestFeatureValidator(TestDataMixin, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.pr_num = 73

        self.valid_feature_str = dedent(
            '''
            from sklearn.base import BaseEstimator, TransformerMixin
            class IdentityTransformer(BaseEstimator, TransformerMixin):
                def fit(self, X, y=None, **fit_kwargs):
                    return self
                def transform(self, X, **transform_kwargs):
                    return X
            input = 'size'
            transformer = IdentityTransformer()
            '''
        )
        self.invalid_feature_str = dedent(
            '''
            from sklearn.base import BaseEstimator, TransformerMixin
            class RaisingTransformer(BaseEstimator, TransformerMixin):
                def fit(self, X, y=None, **fit_kwargs):
                    raise RuntimeError
                def transform(self, X, **transform_kwargs):
                    raise RuntimeError
            input = 'size'
            transformer = RaisingTransformer()
            '''
        )

    @funcy.contextmanager
    def null_prfv(self):
        with mock_repo() as repo:
            commit_range = 'HEAD^..HEAD'
            contrib_module_path = None
            X = None
            y = None

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(self.pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars):
                yield PullRequestFeatureValidator(
                    repo, self.pr_num, contrib_module_path, X, y)

    def test_prfv_init(self):
        with self.null_prfv() as validator:
            self.assertIsInstance(
                validator.differ, TravisPullRequestBuildDiffer)

    def test_prfv_required_method_ordering(self):
        with self.null_prfv() as validator:
            with self.assertRaises(UnexpectedValidationStateError):
                validator._categorize_file_diffs()
            with self.assertRaises(UnexpectedValidationStateError):
                validator._validate_files()
            with self.assertRaises(UnexpectedValidationStateError):
                validator._collect_features()
            with self.assertRaises(UnexpectedValidationStateError):
                validator._validate_features()
            with self.assertRaises(UnexpectedValidationStateError):
                validator._determine_validation_result()

    def test_prfv_collect_file_diffs(self):
        n = 10
        with mock_repo() as repo:
            commits = make_mock_commits(repo, n=n)
            contrib_module_path = None
            X = None
            y = None
            commit_range = get_diff_str_from_commits(
                commits[0], commits[-1])

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(self.pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars):
                validator = PullRequestFeatureValidator(
                    repo, self.pr_num, contrib_module_path, X, y)
                validator._collect_file_diffs()

                # checks on file_diffs
                self.assertEqual(len(validator.file_diffs), n - 1)
                for diff in validator.file_diffs:
                    self.assertEqual(diff.change_type, 'A')
                    self.assertTrue(diff.b_path.startswith('file'))
                    self.assertTrue(diff.b_path.endswith('.py'))

    @funcy.contextmanager
    def mock_project(self, path_content):
        with mock_repo() as repo:
            for path, content in path_content:
                make_mock_commit(repo, path=path, content=content)
            yield repo

    @funcy.contextmanager
    def _test_prfv_end_to_end(self, path_content, contrib_module_path):
        with self.mock_project(path_content) as repo:
            travis_build_dir = repo.working_tree_dir
            travis_pull_request = str(self.pr_num)
            travis_commit_range = get_diff_str_from_commits(
                repo.head.commit.parents[0], repo.head.commit)
            X, y = self.X, self.y

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': travis_build_dir,
                'TRAVIS_PULL_REQUEST': travis_pull_request,
                'TRAVIS_COMMIT_RANGE': travis_commit_range,
            }
            with patch.dict('os.environ', travis_env_vars):
                yield PullRequestFeatureValidator(
                    repo, self.pr_num, contrib_module_path, X, y)

    def test_prfv_end_to_end_failure_no_features_found(self):
        path_content = [
            ('readme.txt', None),
            ('src/__init__.py', None),
            ('src/contrib/__init__.py', None),
            ('src/contrib/foo.py', None),
            ('src/contrib/baz.py', None),
        ]
        contrib_module_path = 'src/contrib/'
        with self._test_prfv_end_to_end(path_content, contrib_module_path) \
                as validator:
            result = validator.validate()
            self.assertFalse(result)

    def test_prfv_end_to_end_failure_inadmissible_file_diffs(self):
        path_content = [
            ('readme.txt', None),
            ('src/__init__.py', None),
            ('src/contrib/__init__.py', None),
            ('src/contrib/foo.py', None),
            ('invalid.py', None),
        ]
        contrib_module_path = 'src/contrib/'
        with self._test_prfv_end_to_end(path_content, contrib_module_path) \
                as validator:
            result = validator.validate()
            self.assertFalse(result)
            self.assertEqual(
                len(validator.file_diffs), 1)
            self.assertEqual(
                len(validator.file_diffs_admissible), 0)
            self.assertEqual(
                len(validator.file_diffs_inadmissible), 1)
            self.assertEqual(
                validator.file_diffs_inadmissible[0].b_path, 'invalid.py')
            self.assertFalse(
                validator.file_diffs_validation_result)

    def test_prfv_end_to_end_failure_bad_package_structure(self):
        path_content = [
            ('foo.jpg', None),
            ('src/contrib/bar/baz.py', self.valid_feature_str),
        ]
        contrib_module_path = 'src/contrib/'
        with self._test_prfv_end_to_end(path_content, contrib_module_path) \
                as validator:
            result = validator.validate()
            self.assertFalse(result)
            self.assertEqual(
                len(validator.file_diffs), 1)
            self.assertEqual(
                len(validator.file_diffs_admissible), 1)
            self.assertEqual(
                len(validator.file_diffs_inadmissible), 0)
            self.assertTrue(
                validator.file_diffs_validation_result)
            self.assertEqual(
                len(validator.features), 0)
            self.assertFalse(
                validator.features_validation_result)

    def test_prfv_end_to_end_failure_invalid_feature(self):
        path_content = [
            ('foo.jpg', None),
            ('src/__init__.py', None),
            ('src/contrib/__init__.py', None),
            ('src/contrib/foo.py', self.invalid_feature_str),
        ]
        contrib_module_path = 'src/contrib/'
        with self._test_prfv_end_to_end(path_content, contrib_module_path) \
                as validator:
            result = validator.validate()
            self.assertFalse(result)
            self.assertEqual(
                len(validator.file_diffs), 1)
            self.assertEqual(
                len(validator.file_diffs_admissible), 1)
            self.assertEqual(
                len(validator.file_diffs_inadmissible), 0)
            self.assertTrue(
                validator.file_diffs_validation_result)
            self.assertEqual(
                len(validator.features), 1)
            self.assertFalse(
                validator.features_validation_result)

    def test_prfv_end_to_end_success(self):
        path_content = [
            ('bob.xml', '<><> :: :)'),
            ('src/__init__.py', None),
            ('src/contrib/__init__.py', None),
            ('src/contrib/bean.py', self.valid_feature_str),
        ]
        contrib_module_path = 'src/contrib/'
        with self._test_prfv_end_to_end(path_content, contrib_module_path) \
                as validator:
            result = validator.validate()
            self.assertTrue(result)



# -*- coding: utf-8 -*-

"""Top-level package for ballet."""

__author__ = """Micah Smith"""
__email__ = 'micahs@mit.edu'
__version__ = '0.4.1'

# pathlib compatibility
import sys
if sys.version_info < (3, 5, 0):
    import pathlib2 as pathlib
else:
    import pathlib  # noqa F401


# re-export some names
from ballet.feature import *  # noqa
from ballet.contrib import *  # noqa


# configure module-level logging
import logging  # noqa E402
from ballet.util.log import logger  # noqa E402
logger.addHandler(logging.NullHandler())



import unittest

import numpy as np
import pandas as pd
import sklearn.base

import ballet.eng.base
import ballet.eng.misc


class TestBase(unittest.TestCase):
    def setUp(self):
        pass

    def test_no_fit_mixin(self):
        class _Foo(ballet.eng.base.NoFitMixin):
            pass
        a = _Foo()

        self.assertTrue(hasattr(a, 'fit'))

        # method should exist
        a.fit('X')
        a.fit('X', y=None)


    def test_base_transformer(self):
        a = ballet.eng.base.BaseTransformer()

        self.assertIsInstance(a, sklearn.base.BaseEstimator)
        self.assertTrue(hasattr(a, 'fit'))

    def test_simple_function_transformer(self):
        def func(x): return x + 5
        data = np.arange(30)

        trans = ballet.eng.base.SimpleFunctionTransformer(func)
        trans.fit(data)
        data_trans = trans.transform(data)
        data_func = func(data)

        self.assertTrue(np.array_equal(data_trans, data_func))

    def test_grouped_function_transformer(self):
        df = pd.DataFrame(
            data={
                'country': ['USA', 'USA', 'USA', 'Canada', 'Fiji'],
                'year': [2001, 2002, 2003, 2001, 2001],
                'length': [1, 2, 3, 4, 5],
                'width': [1.0, 1.0, 7.5, 9.0, 11.0],
            }
        ).set_index(['country', 'year']).sort_index()

        # with groupby kwargs, produces a df
        func = np.sum
        trans = ballet.eng.base.GroupedFunctionTransformer(
            func, groupby_kwargs={'level': 'country'})
        trans.fit(df)
        result = trans.transform(df)
        expected_result = df.groupby(level='country').apply(func)
        pd.util.testing.assert_frame_equal(result, expected_result)

        # without groupby kwargs, produces a series
        func = np.min
        trans = ballet.eng.base.GroupedFunctionTransformer(func)
        trans.fit(df)
        result = trans.transform(df)
        expected_result = df.pipe(func)
        pd.util.testing.assert_series_equal(result, expected_result)


class TestMisc(unittest.TestCase):
    def setup(self):
        pass

    def test_value_replacer(self):
        trans = ballet.eng.misc.ValueReplacer(0.0, -99)
        data = pd.DataFrame([0, 0, 0, 0, 1, 3, 7, 11, -7])
        expected_result = pd.DataFrame([-99, -99, -99, -99, 1, 3, 7, 11, -7])

        result = trans.fit_transform(data)
        pd.util.testing.assert_frame_equal(result, expected_result)

    def test_named_framer(self):
        name = 'foo'

        # good objects
        index = pd.Index([1, 2, 3])
        ser = pd.Series([1, 2, 3])
        df = pd.DataFrame([1, 2, 3])
        arr = np.array([1, 2, 3])

        for obj in [index, ser, df, arr]:
            trans = ballet.eng.misc.NamedFramer(name)
            result = trans.fit_transform(obj)
            self.assertTrue(isinstance(result, pd.DataFrame))
            self.assertEqual(result.shape[1], 1)
            self.assertEqual(result.columns, [name])

        # bad objects -- wrong type
        int_ = 1
        str_ = 'hello'
        for obj in [int_, str_]:
            trans = ballet.eng.misc.NamedFramer(name)
            with self.assertRaises(TypeError):
                result = trans.fit_transform(obj)

        # bad objects -- too wide
        wide_df = pd.DataFrame(np.arange(10).reshape(-1, 2))
        wide_arr = np.arange(10).reshape(-1, 2)
        for obj in [wide_df, wide_arr]:
            trans = ballet.eng.misc.NamedFramer(name)
            with self.assertRaises(ValueError):
                result = trans.fit_transform(obj)



RANDOM_STATE = 1754


def _make_multiclass(lst):
    return [
        l + '_' + MULTICLASS_METRIC_AGGREGATION
        for l in lst
    ]


MULTICLASS_METRIC_AGGREGATION = "micro"

CLASSIFICATION_SCORING = ['accuracy', 'roc_auc']

BINARY_CLASSIFICATION_SCORING = (
    CLASSIFICATION_SCORING + ['precision', 'recall'])

MULTICLASS_CLASSIFICATION_SCORING = (
    CLASSIFICATION_SCORING + _make_multiclass(['precision', 'recall'])
)

REGRESSION_SCORING = ['neg_mean_squared_error', 'r2']

SCORING_NAME_MAPPER = {
    # classification
    'accuracy': 'Accuracy',
    'average_precision': 'Average Precision',
    'f1': 'F1 Score (Binary)',
    'f1_micro': 'F1 Score (micro-averaged)',
    'f1_macro': 'F1 Score (macro-averaged)',
    'f1_weighted': 'F1 Score (weighted average)',
    'f1_samples': 'F1 Score (by multilabel sample)',
    'neg_log_loss': 'Negative Log Loss',
    'precision': 'Precision (Binary)',
    'precision_micro': 'Precision (micro-averaged)',
    'precision_macro': 'Precision (macro-averaged)',
    'precision_weighted': 'Precision (weighted average)',
    'precision_samples': 'Precision (by multilabel sample)',
    'recall': 'Recall (Binary)',
    'recall_micro': 'Recall (micro-averaged)',
    'recall_macro': 'Recall (macro-averaged)',
    'recall_weighted': 'Recall (weighted average)',
    'recall_samples': 'Recall (by multilabel sample)',
    'roc_auc': 'ROC AUC Score',

    # regression
    'explained_variance': 'Explained Variance',
    'neg_mean_absolute_error': 'Negative Mean Absolute Error',
    'neg_mean_squared_error': 'Negative Mean Squared Error',
    'neg_mean_squared_lod_error': 'Negative Mean Squared Log Error',
    'neg_median_absolute_error': 'Negative Median Absolute Error',
    'r2': 'R-squared',
}



import traceback

import funcy
import numpy as np
import pandas as pd
from sklearn.pipeline import _name_estimators
from sklearn_pandas import DataFrameMapper
from sklearn_pandas.pipeline import TransformerPipeline

from ballet.util import asarray2d, indent
from ballet.util.log import logger

__all__ = ['make_mapper', 'Feature']


def make_mapper(features):
    """Make a DataFrameMapper from a list of features

    Args:
        features (List[Feature]): list of features

    Returns:
        DataFrameMapper: mapper made from features
    """
    return DataFrameMapper(
        [t.as_input_transformer_tuple() for t in features],
        input_df=True)


class RobustTransformerPipeline(TransformerPipeline):

    def transform(self, X, **transform_kwargs):
        _transform = make_robust_to_tabular_types(super().transform)
        return _transform(X, **transform_kwargs)


def make_robust_transformer_pipeline(*steps):
    '''Construct a RobustTransformerPipeline from the given estimators.'''
    return RobustTransformerPipeline(_name_estimators(steps))


def make_conversion_approaches():
    funcs = (funcy.identity, pd.Series, pd.DataFrame, np.asarray, asarray2d)
    catch = (ValueError, TypeError)
    for func in funcs[:-1]:
        yield func, catch
    yield funcs[-1], ()


def make_robust_to_tabular_types(func):
    @funcy.wraps(func)
    def wrapped(X, y=None, **kwargs):
        for convert, catch in make_conversion_approaches():
            try:
                logger.debug(
                    "Converting using approach '{}'".format(convert.__name__))
                if y is not None:
                    return func(convert(X), y=convert(y), **kwargs)
                else:
                    return func(convert(X), **kwargs)
            except catch as e:
                formatted_exc = indent(traceback.format_exc(), n=8)
                logger.debug(
                    "Application subsequently failed with exception '{}'\n\n{}"
                    .format(e.__class__.__name__, formatted_exc))
    return wrapped


def make_robust_transformer(transformer):
    transformer.fit = make_robust_to_tabular_types(transformer.fit)
    transformer.transform = make_robust_to_tabular_types(transformer.transform)
    return transformer


class Feature:
    def __init__(self, input, transformer, name=None, description=None,
                 output=None, source=None, options=None):
        self.input = input
        if funcy.is_seqcont(transformer):
            transformer = make_robust_transformer_pipeline(*transformer)
        self.transformer = make_robust_transformer(transformer)
        self.name = name
        self.description = description
        self.output = output  # unused
        self.source = source
        self.options = options if options is not None else {}

    def __repr__(self):
        # TODO use self.__dict__ directly, which respects insertion order
        attr_list = ['input', 'transformer', 'name', 'description', 'output',
                     'source', 'options']
        attrs_str = ', '.join(
            '{attr_name}={attr_val!r}'.format(
                attr_name=attr, attr_val=getattr(self, attr)
            ) for attr in attr_list
        )
        return self.__class__.__name__ + '(' + attrs_str + ')'

    def as_input_transformer_tuple(self):
        return (self.input, self.transformer)

    def as_dataframe_mapper(self):
        return DataFrameMapper([
            self.as_input_transformer_tuple()
        ], input_df=True)



import os
import pathlib
import tempfile
import types
import unittest
from unittest.mock import patch

import ballet
from ballet.util.gitutil import get_diff_str_from_commits
from ballet.util.modutil import (  # noqa F401
    import_module_at_path, import_module_from_modname,
    import_module_from_relpath, modname_to_relpath, relpath_to_modname)
from ballet.util.travisutil import (
    TravisPullRequestBuildDiffer, get_travis_pr_num, is_travis_pr)

from .util import make_mock_commits, mock_repo


class TestModutil(unittest.TestCase):

    @unittest.expectedFailure
    def test_import_module_from_modname(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_import_module_from_relpath(self):
        raise NotImplementedError

    def test_import_module_at_path_module(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir).joinpath('foo', 'bar.py')
            path.parent.mkdir(parents=True)
            init = path.parent.joinpath('__init__.py')
            init.touch()
            x = 1
            with path.open('w') as f:
                f.write('x={x!r}'.format(x=x))
            modname = 'foo.bar'
            modpath = str(path)  # e.g. /tmp/foo/bar.py'
            mod = import_module_at_path(modname, modpath)
            self.assertIsInstance(mod, types.ModuleType)
            self.assertEqual(mod.__name__, modname)
            self.assertEqual(mod.x, x)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir).joinpath('foo')
            path.mkdir(parents=True)
            init = path.joinpath('__init__.py')
            init.touch()
            x = 'hello'
            with init.open('w') as f:
                f.write('x={x!r}'.format(x=x))
            modname = 'foo'
            modpath = str(path)
            mod = import_module_at_path(modname, modpath)
            self.assertIsInstance(mod, types.ModuleType)
            self.assertEqual(mod.__name__, modname)
            self.assertEqual(mod.x, x)

    @unittest.expectedFailure
    def test_import_module_at_path_bad_package_structure(self):
        raise NotImplementedError

    def test_relpath_to_modname(self):
        relpath = 'ballet/util/_util.py'
        expected_modname = 'ballet.util._util'
        actual_modname = relpath_to_modname(relpath)
        self.assertEqual(actual_modname, expected_modname)

        relpath = 'ballet/util/__init__.py'
        expected_modname = 'ballet.util'
        actual_modname = relpath_to_modname(relpath)
        self.assertEqual(actual_modname, expected_modname)

        relpath = 'ballet/foo/bar/baz.zip'
        with self.assertRaises(ValueError):
            relpath_to_modname(relpath)

    def test_modname_to_relpath(self):
        modname = 'ballet.util._util'
        expected_relpath = 'ballet/util/_util.py'
        actual_relpath = modname_to_relpath(modname)
        self.assertEqual(actual_relpath, expected_relpath)

        modname = 'ballet.util'
        # mypackage.__file__ resolves to the __init__.py
        project_root = pathlib.Path(ballet.__file__).parent.parent

        expected_relpath = 'ballet/util/__init__.py'
        actual_relpath = modname_to_relpath(modname, project_root=project_root)
        self.assertEqual(actual_relpath, expected_relpath)

        # without providing project root, behavior is undefined, as we don't
        # know whether the relative path will resolve to a directory

        # within a temporary directory, the relpath *should not* be a dir
        with tempfile.TemporaryDirectory() as tmpdir:
            cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                actual_relpath = modname_to_relpath(modname)
                expected_relpath = 'ballet/util.py'
                self.assertEqual(actual_relpath, expected_relpath)
            finally:
                os.chdir(cwd)

        # from the actual project root, the relpath *should* be a dir
        cwd = os.getcwd()
        try:
            os.chdir(str(project_root))
            actual_relpath = modname_to_relpath(modname)
            expected_relpath = 'ballet/util/__init__.py'
            self.assertEqual(actual_relpath, expected_relpath)
        finally:
            os.chdir(cwd)

        # without add_init
        modname = 'ballet.util'
        add_init = False
        expected_relpath = 'ballet/util'
        actual_relpath = modname_to_relpath(
            modname, project_root=project_root, add_init=add_init)
        self.assertEqual(actual_relpath, expected_relpath)


class TestTravis(unittest.TestCase):
    def setUp(self):
        self.pr_num = 7
        self.commit_range = 'HEAD^..HEAD'
        self.travis_vars = {
            'TRAVIS_PULL_REQUEST': str(self.pr_num),
            'TRAVIS_COMMIT_RANGE': self.commit_range,
        }

    def test_get_travis_pr_num(self):
        # matrix of env name, setting for env, expected result
        matrix = (
            ('TRAVIS_PULL_REQUEST', str(self.pr_num), self.pr_num),
            ('TRAVIS_PULL_REQUEST', 'true', None),
            ('TRAVIS_PULL_REQUEST', 'FALSE', None),
            ('TRAVIS_PULL_REQUEST', 'false', None),
            ('TRAVIS_PULL_REQUEST', 'abcd', None),
            ('UNRELATED_VAR', '', None),
        )
        for env_name, env_value, expected_result in matrix:
            with patch.dict('os.environ', {env_name: env_value}):
                actual_result = get_travis_pr_num()
                self.assertEqual(actual_result, expected_result)

    def test_is_travis_pr(self):
        matrix = (
            ('TRAVIS_PULL_REQUEST', str(self.pr_num), True),
            ('TRAVIS_PULL_REQUEST', 'true', False),
            ('TRAVIS_PULL_REQUEST', 'FALSE', False),
            ('TRAVIS_PULL_REQUEST', 'false', False),
            ('TRAVIS_PULL_REQUEST', 'abcd', False),
            ('UNRELATED_VAR', '', False),
        )
        for env_name, env_value, expected_result in matrix:
            with patch.dict('os.environ', {env_name: env_value}):
                actual_result = is_travis_pr()
                self.assertEqual(actual_result, expected_result)

    def test_travis_pull_request_build_differ(self):
        with mock_repo() as repo:
            pr_num = self.pr_num
            commit_range = 'HEAD^..HEAD'

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars):
                differ = TravisPullRequestBuildDiffer(pr_num)
                diff_str = differ._get_diff_str()
                self.assertEqual(diff_str, commit_range)

    def test_travis_pull_request_build_differ_on_mock_commits(self):
        n = 10
        i = 0
        pr_num = self.pr_num
        with mock_repo() as repo:
            commits = make_mock_commits(repo, n=n)
            commit_range = get_diff_str_from_commits(commits[i], commits[-1])

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars):
                differ = TravisPullRequestBuildDiffer(pr_num)
                diff_str = differ._get_diff_str()
                self.assertEqual(diff_str, commit_range)

                diffs = differ.diff()

                # there should be n-1 diff objects, they should show files
                # 1 to n-1
                self.assertEqual(len(diffs), n - 1)
                j = i + 1
                for diff in diffs:
                    self.assertEqual(diff.change_type, 'A')
                    self.assertEqual(diff.b_path, 'file{j}.py'.format(j=j))
                    j += 1



import unittest

import numpy as np
import pandas as pd
import pandas.util.testing as pdt

from ballet.util.testing import ArrayLikeEqualityTestingMixin


class TestArrayLikeEqualityTestingMixin(
        ArrayLikeEqualityTestingMixin, unittest.TestCase):

    def setUp(self):
        pass

    def test_assert_array_equal(self):
        a = np.arange(10 * 7).reshape(10, 7)
        b = a.copy()
        self.assertArrayEqual(a, b)

        c = np.arange(9).reshape(3, 3)
        d = c + 1

        with self.assertRaises(AssertionError):
            self.assertArrayEqual(c, d)

    def test_assert_array_not_equal(self):
        a = np.arange(10 * 7).reshape(10, 7)
        b = a.copy()

        with self.assertRaises(AssertionError):
            self.assertArrayNotEqual(a, b)

        c = np.arange(9).reshape(3, 3)
        d = c + 1

        self.assertArrayNotEqual(c, d)

    def test_assert_frame_equal(self):
        a = pdt.makeCustomDataframe(
            10, 7, data_gen_f=lambda row, col: row * col)
        b = a.copy()
        self.assertFrameEqual(a, b)

        c = a + 1
        with self.assertRaises(
                AssertionError, msg='Should find that frames are same size '
                                    'but different values'):
            self.assertFrameEqual(a, c)

        d = pdt.makeCustomDataframe(11, 9)
        with self.assertRaises(
                AssertionError, msg='Should find that frames are different '
                                    'size'):
            self.assertFrameEqual(a, d)

        e = pdt.makeTimeSeries()
        with self.assertRaises(
                AssertionError, msg='Should find that the second arg is '
                                    'wrong type'):
            self.assertFrameEqual(a, e)

        f = pd.DataFrame([1, 2, 3, 4])
        g = pd.Series([1, 2, 3, 4])
        with self.assertRaises(
                AssertionError,
                msg='Should find that the second arg is wrong type, '
                    'even though they have the same values'):
            self.assertFrameEqual(f, g)

    def test_assert_frame_not_equal(self):
        a = pdt.makeCustomDataframe(
            10, 7, data_gen_f=lambda row, col: row * col)
        b = a.copy()
        with self.assertRaises(AssertionError,
                               msg='frames are actually equal'):
            self.assertFrameNotEqual(a, b)

        c = a + 1
        self.assertFrameNotEqual(a, c)

        d = pdt.makeCustomDataframe(11, 9)
        self.assertFrameNotEqual(a, d)

        e = pdt.makeTimeSeries()
        self.assertFrameNotEqual(a, e)

        f = pd.DataFrame([1, 2, 3, 4])
        g = pd.Series([1, 2, 3, 4])
        self.assertFrameNotEqual(f, g)

    def test_assert_series_equal(self):
        a = pd.Series(np.arange(21))
        b = a.copy()
        self.assertSeriesEqual(a, b)

        c = a + 1
        with self.assertRaises(
                AssertionError, msg='Should find that series are same size '
                                    'but different values'):
            self.assertSeriesEqual(a, c)

        d = pd.Series(np.arange(17))
        with self.assertRaises(
                AssertionError, msg='Should find that series are different '
                                    ' sizes'):
            self.assertSeriesEqual(a, d)

        e = pdt.makeDataFrame()
        with self.assertRaises(
                AssertionError, msg='Should find that the second arg is '
                                    'wrong type'):
            self.assertSeriesEqual(a, e)

    def test_assert_series_not_equal(self):
        a = pd.Series(np.arange(21))
        b = a.copy()
        with self.assertRaises(
                AssertionError, msg='Should find that the series are '
                                    'indeed equal'):
            self.assertSeriesNotEqual(a, b)

        c = a + 1
        self.assertSeriesNotEqual(a, c)

        d = pd.Series(np.arange(17))
        self.assertSeriesNotEqual(a, d)

        e = pdt.makeDataFrame()
        self.assertSeriesNotEqual(a, e)

    def test_assert_index_equal(self):
        a = pd.Index(np.arange(21))
        b = a.copy()
        self.assertIndexEqual(a, b)

        c = a + 1
        with self.assertRaises(
                AssertionError, msg='Should find that indexes are same '
                                    'size but different values'):
            self.assertIndexEqual(a, c)

        d = pd.Index(np.arange(17))
        with self.assertRaises(
                AssertionError, msg='Should find that indexes are '
                                    'different sizes'):
            self.assertIndexEqual(a, d)

        e = pdt.makeDataFrame()
        with self.assertRaises(
                AssertionError, msg='Should find that the second arg is '
                                    'wrong type'):
            self.assertIndexEqual(a, e)

    def test_assert_index_not_equal(self):
        a = pd.Index(np.arange(21))
        b = a.copy()
        with self.assertRaises(
                AssertionError, msg='Should find that indexes are '
                                    'indeed equal'):
            self.assertIndexNotEqual(a, b)

        c = a + 1
        self.assertIndexNotEqual(a, c)

        d = pd.Index(np.arange(17))
        self.assertIndexNotEqual(a, d)

        e = pdt.makeDataFrame()
        self.assertIndexNotEqual(a, e)



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import LabelBinarizer


class FeatureTypeTransformer(BaseEstimator, TransformerMixin):
    BAD_TYPE_MSG = "Unsupported input type '{}'"
    BAD_SHAPE_MSG = "Unsupported input shape '{}'"

    @staticmethod
    def _get_original_info(X):
        if isinstance(X, pd.Series):
            return {
                'index': X.index,
                'dtype': X.dtype,
                'name': X.name,
            }
        elif isinstance(X, pd.DataFrame):
            return {
                'index': X.index,
                'dtypes': X.dtypes,
                'columns': X.columns,
            }
        elif isinstance(X, np.ndarray):
            return {'ndim': X.ndim}
        else:
            return {}

    def fit(self, X, **fit_kwargs):
        self.original_type_ = type(X)
        self.original_info_ = self._get_original_info(X)
        return self

    def transform(self, X, **transform_kwargs):
        if not hasattr(self, 'original_type_'):
            raise NotFittedError

        if isinstance(X, pd.Series):
            return X.values
            # return X.to_frame().to_records(index=False))
        elif isinstance(X, pd.DataFrame):
            return X.values
            # return np.asarray(X.to_records(index=False))
        elif isinstance(X, np.ndarray):
            if X.ndim == 1:
                return X.reshape(-1, 1)
            elif X.ndim == 2:
                return X
            elif X.ndim >= 3:
                raise TypeError(
                    FeatureTypeTransformer.BAD_SHAPE_MSG.format(
                        X.shape))
        else:
            # should be unreachable
            raise TypeError(
                FeatureTypeTransformer.BAD_TYPE_MSG.format(
                    type(X)))

    def inverse_transform(self, X, **inverse_transform_kwargs):
        if not hasattr(self, 'original_type_'):
            raise NotFittedError

        if hasattr(self, 'original_type_') and hasattr(self, 'original_info_'):
            if issubclass(self.original_type_, pd.Series):
                data = X
                index = self.original_info_['index']
                name = self.original_info_['name']
                dtype = self.original_info_['dtype']
                return pd.Series(data=data, index=index,
                                 name=name, dtype=dtype)
            elif issubclass(self.original_type_, pd.DataFrame):
                data = X
                index = self.original_info_['index']
                columns = self.original_info_['columns']
                dtypes = self.original_info_['dtypes']
                df = pd.DataFrame(data=data, index=index,
                                  columns=columns)
                df = df.astype(dtype=dtypes.to_dict())
                return df
            elif issubclass(self.original_type_, np.ndarray):
                # only thing we might have done is change dimensions for 1d/2d
                if self.original_info_['ndim'] == 1:
                    return X.ravel()
                elif self.original_info_['ndim'] == 2:
                    return X
            else:
                # unreachable
                raise NotImplementedError
        else:
            raise NotFittedError


class TargetTypeTransformer(FeatureTypeTransformer):
    def __init__(self, needs_label_binarizer=False):
        super().__init__()
        self.needs_label_binarizer = needs_label_binarizer

    def fit(self, y, **fit_kwargs):
        super().fit(y, **fit_kwargs)
        if self.needs_label_binarizer:
            self.label_binarizer_ = LabelBinarizer()
            self.label_binarizer_.fit(y)
        return self

    def transform(self, y, **transform_kwargs):
        y = super().transform(y)
        if self.needs_label_binarizer:
            y = self.label_binarizer_.transform(y)
        else:
            y = y.ravel()
        return y

    def inverse_transform(self, y, **inverse_transform_kwargs):
        if self.needs_label_binarizer:
            y = self.label_binarizer_.inverse_transform(y)
        y = super().inverse_transform(y)
        return y



import unittest

import numpy as np
import pandas as pd
import sklearn.datasets
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import LabelBinarizer

from ballet.modeler import (
    DecisionTreeModeler, StratifiedKFoldMultiClassIndicator, TunedModeler,
    TunedRandomForestClassifier, TunedRandomForestRegressor)
from ballet.modeling.problem import (
    MulticlassClassificationProblem, RegressionProblem)
from ballet.util.log import logger
from ballet.util.testing import EPSILON, log_seed_on_error, seeded


class _CommonTesting:

    def setUp(self):
        # Create fake data
        X_classification, y_classification = sklearn.datasets.load_iris(
            return_X_y=True)
        X_regression, y_regression = sklearn.datasets.load_boston(
            return_X_y=True)

        # TODO: add test for BinaryClassificationProblem
        self.data = {
            MulticlassClassificationProblem: {
                "X": X_classification,
                "y": y_classification,
            },
            RegressionProblem: {
                "X": X_regression,
                "y": y_regression,
            },
        }

        self.data_pd = {
            MulticlassClassificationProblem: {
                "X": pd.DataFrame(X_classification),
                "y": pd.DataFrame(y_classification),
            },
            RegressionProblem: {
                "X": pd.DataFrame(X_regression),
                "y": pd.DataFrame(y_regression),
            },
        }

    def _setup_modeler(self, problem_type, data):
        X = data[problem_type]["X"]
        y = data[problem_type]["y"]
        modeler = self.modeler_class(problem_type=problem_type)
        return modeler, X, y

    def _test_problem_type_cv(self, problem_type, data):
        modeler, X, y = self._setup_modeler(problem_type, data)
        metrics = modeler.compute_metrics_cv(X, y)
        return metrics

    def _test_problem_type_train_test(self, problem_type, data):
        modeler, X, y = self._setup_modeler(problem_type, data)
        n = round(0.7 * len(X))
        metrics = modeler.compute_metrics_train_test(X, y, n=n)

        return metrics

    def _call_method(self, method, problem_type, seed=None):
        with seeded(seed):
            metrics = getattr(self, method)(problem_type, self.data)
        with seeded(seed):
            metrics_pd = getattr(self, method)(problem_type, self.data_pd)
        return metrics, metrics_pd

    def _prepare_metrics_for_assertions(self, metrics):
        return {
            metric['description']: metric['value']
            for metric in metrics
        }


class TestModeler(_CommonTesting, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.modeler_class = DecisionTreeModeler

    def test_classification_cv(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_cv', MulticlassClassificationProblem, seed=17)
        self.assertEqual(metrics, metrics_pd)
        metrics = self._prepare_metrics_for_assertions(metrics)
        self.assertAlmostEqual(
            metrics['Accuracy'], 0.9, delta=0.1 - EPSILON)
        self.assertAlmostEqual(
            metrics['ROC AUC Score'], 0.9, delta=0.1 - EPSILON)

        # todo multiclass
        # self.assertAlmostEqual(
        #     metrics['Precision'], 0.9403594, delta=EPSILON)
        # self.assertAlmostEqual(
        #     metrics['Recall'], 0.9403594, delta=EPSILON)

    def test_classification_train_test(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_train_test', MulticlassClassificationProblem,
            seed=1093)
        self.assertEqual(metrics, metrics_pd)
        metrics = self._prepare_metrics_for_assertions(metrics)
        self.assertAlmostEqual(
            metrics['Accuracy'], 0.9, delta=0.1 - EPSILON)
        self.assertAlmostEqual(
            metrics['ROC AUC Score'], 0.9, delta=0.1 - EPSILON)

        # # todo multiclass
        # self.assertAlmostEqual(
        #     metrics['Precision'], 0.7777777, delta=EPSILON)
        # self.assertAlmostEqual(
        #     metrics['Recall'], 0.7777777, delta=EPSILON)

    def test_regression_cv(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_cv', RegressionProblem)
        self.assertEqual(metrics, metrics_pd)
        metrics = self._prepare_metrics_for_assertions(metrics)
        self.assertAlmostEqual(
            metrics['Negative Mean Squared Error'], -20.7, delta=0.1 - EPSILON)
        self.assertAlmostEqual(
            metrics['R-squared'], 0.7, delta=0.1 - EPSILON)

    def test_regression_train_test(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_train_test', RegressionProblem, seed=4)
        self.assertEqual(metrics, metrics_pd)
        metrics = self._prepare_metrics_for_assertions(metrics)
        self.assertAlmostEqual(
            metrics['Negative Mean Squared Error'], -34.7, delta=0.1 - EPSILON)
        self.assertAlmostEqual(
            metrics['R-squared'], 0.6, delta=0.1 - EPSILON)


class TestTunedModelers(_CommonTesting, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.modeler_class = TunedModeler

    def _setup_modeler(self, problem_type, data):
        modeler, X, y = super()._setup_modeler(problem_type, data)
        return modeler, X, y

    def test_classification_cv(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_cv', MulticlassClassificationProblem, seed=1)
        self.assertEqual(metrics, metrics_pd)

    def test_classification_train_test(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_train_test', MulticlassClassificationProblem,
            seed=2)
        self.assertEqual(metrics, metrics_pd)

    def test_regression_cv(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_cv', RegressionProblem, seed=3)

    def test_regression_train_test(self):
        metrics, metrics_pd = self._call_method(
            '_test_problem_type_train_test', RegressionProblem, seed=4)
        self.assertEqual(metrics, metrics_pd)

    def _test_tuned_random_forest_estimator(self, Estimator, problem_type):
        model = Estimator()
        data = self.data[problem_type]
        X, y = data['X'], data['y']
        with log_seed_on_error(logger):
            model.fit(X, y, tune=False)
            old_score = model.score(X, y)
            model.fit(X, y, tune=True)
            new_score = model.score(X, y)
            self.assertGreaterEqual(new_score, old_score)

    def test_tuned_random_forest_regressor(self):
        self._test_tuned_random_forest_estimator(
            TunedRandomForestRegressor, RegressionProblem)

    def test_tuned_random_forest_classifier(self):
        self._test_tuned_random_forest_estimator(
            TunedRandomForestClassifier, MulticlassClassificationProblem)


class TestStratifiedKFoldMultiClassIndicator(unittest.TestCase):
    def test(self):
        X, y = sklearn.datasets.load_iris(return_X_y=True)
        transformer = LabelBinarizer()
        ym = transformer.fit_transform(y)

        params = {'random_state': 1, 'n_splits': 3}
        kf = StratifiedKFold(**params)
        kfm = StratifiedKFoldMultiClassIndicator(transformer, **params)

        kf_folds = kf.split(X, y)
        kfm_folds = kfm.split(X, ym)

        for (inds_tr_kf, inds_te_kf), (inds_tr_kfm, inds_te_kfm) \
                in zip(kf_folds, kfm_folds):
            self.assertTrue(np.array_equal(inds_tr_kf, inds_tr_kfm))
            self.assertTrue(np.array_equal(inds_te_kf, inds_te_kfm))



import os
import random
import tempfile

import funcy
import git
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn_pandas.pipeline import TransformerPipeline

from ballet import pathlib
from ballet.eng.misc import IdentityTransformer

EPSILON = 1e-4


class FragileTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, bad_input_checks, errors):
        '''Raises a random error if any input check returns True'''
        super().__init__()

        self._check = funcy.any_fn(*bad_input_checks)
        self._errors = errors

        self._random = random.Random()
        self._random.seed = hash(funcy.merge(bad_input_checks, errors))

    def _raise(self):
        raise self._random.choice(self._errors)

    def fit(self, X, y=None, **fit_kwargs):
        if self._check(X) or self._check(y):
            self._raise()

        return self

    def transform(self, X, **transform_kwargs):
        if self._check(X):
            self._raise()

        return X


class FragileTransformerPipeline(TransformerPipeline):
    def __init__(self, nsteps, bad_input_checks, errors, shuffle=True, seed=1):
        steps = [
            ('IdentityTransformer{:02d}'.format(i), IdentityTransformer())
            for i in range(nsteps - 1)
        ]
        fragile_transformer = FragileTransformer(bad_input_checks, errors)
        steps.append(
            (repr(fragile_transformer), fragile_transformer)
        )
        if shuffle:
            rand = random.Random()
            rand.seed(seed)
            rand.shuffle(steps)

        super().__init__(steps)


def make_mock_commit(repo, kind='A', path=None, content=None):
    '''Commits one file to repo'''
    if not path:
        path = 'file{}'.format(random.randint(0, 999))

    # TODO resolve relative to root, then delete root
    # path = pathlib.Path(path).resolve()
    # dir = pathlib.Path(repo.working_tree_dir).resolve()
    # if dir not in path.parents:
    #     raise ValueError(
    #         'Path {} must be a relative path to a subdirectory of the '
    #         'repo root.'.format(str(path))

    dir = repo.working_tree_dir
    abspath = pathlib.Path(dir).joinpath(path)
    if kind == 'A':
        # TODO make robust
        abspath.parent.mkdir(parents=True, exist_ok=True)

        if abspath.exists():
            # because this would be a kind=='M'
            raise FileExistsError(str(abspath))
        else:
            if content is not None:
                with abspath.open('w') as f:
                    f.write(content)
            else:
                abspath.touch()
        repo.git.add(str(abspath))
        repo.git.commit(m='Commit {}'.format(str(abspath)))
    else:
        raise NotImplementedError

    return repo.head.commit


def make_mock_commits(repo, n=10):
    '''Create n sequential files/commits'''
    commits = []
    for i in range(n):
        commit = make_mock_commit(repo, path='file{}.py'.format(i))
        commits.append(commit)
    return commits


@funcy.contextmanager
def mock_repo():
    '''Create a new repo'''
    with tempfile.TemporaryDirectory() as tmpdir:
        cwd = os.getcwd()
        os.chdir(str(tmpdir))
        dir = pathlib.Path(tmpdir)
        repo = git.Repo.init(str(dir))
        try:
            yield repo
        finally:
            os.chdir(cwd)



import logging

import ballet

logger = logging.getLogger(ballet.__name__)

LOG_FORMAT = r'[%(asctime)s] {%(name)s: %(filename)s:%(lineno)d} %(levelname)s - %(message)s'  # noqa E501
SIMPLE_LOG_FORMAT = r'%(asctime)s %(levelname)s - %(message)s'


def enable(level=logging.INFO):
    """Enable simple console logging for this module"""
    logger.setLevel(logging.DEBUG)
    stderr_handler = logging.StreamHandler()
    stderr_handler.setLevel(level)
    formatter = logging.Formatter(LOG_FORMAT)
    stderr_handler.setFormatter(formatter)
    logger.addHandler(stderr_handler)
    logger.info('Logging enabled.')



import funcy
from sklearn.base import BaseEstimator, TransformerMixin

__all__ = [
    'BaseTransformer',
    'GroupedFunctionTransformer',
    'NoFitMixin',
    'SimpleFunctionTransformer',
]


class NoFitMixin:
    def fit(self, X, y=None, **fit_kwargs):
        return self


class BaseTransformer(NoFitMixin, TransformerMixin, BaseEstimator):
    pass


class SimpleFunctionTransformer(BaseTransformer):
    def __init__(self, func, func_args=None, func_kwargs=None):
        super().__init__()
        self.func = func
        self.func_args = func_args if func_args else ()
        self.func_kwargs = func_kwargs if func_kwargs else {}
        self.func_call = funcy.rpartial(
            self.func, *self.func_args, **self.func_kwargs)

    def transform(self, X, **transform_kwargs):
        return self.func_call(X)


class GroupedFunctionTransformer(SimpleFunctionTransformer):
    def __init__(self, func, func_args=None,
                 func_kwargs=None, groupby_kwargs=None):
        super().__init__(func, func_args=func_args, func_kwargs=func_kwargs)
        self.groupby_kwargs = groupby_kwargs if groupby_kwargs else {}

    def transform(self, X, **transform_kwargs):
        if self.groupby_kwargs:
            call = X.sort_index().groupby(**self.groupby_kwargs).apply
        else:
            call = X.sort_index().pipe
        return call(self.func_call)


class DelegatingTransformerMixin(TransformerMixin):
    # TODO remove? almost certainly this can be accomplished with inheritance
    # directly
    def fit(self, X, y=None, **fit_args):
        return self._transformer.fit(X, y=None, **fit_args)

    def transform(self, X, **transform_args):
        return self._transformer.transform(X, **transform_args)



import unittest

import funcy
import numpy as np
import pandas as pd
import sklearn.preprocessing
from sklearn_pandas import DataFrameMapper

from ballet.compat import SimpleImputer
from ballet.feature import Feature, make_mapper, make_robust_transformer
from ballet.eng.misc import IdentityTransformer
from ballet.util import asarray2d

from .util import FragileTransformer, FragileTransformerPipeline


class TestFeature(unittest.TestCase):
    def setUp(self):
        self.input = 'foo'

        self.transformer = IdentityTransformer()

        self.X_ser = pd.util.testing.makeFloatSeries()
        self.X_df = self.X_ser.to_frame()
        self.X_arr1d = np.asarray(self.X_ser)
        self.X_arr2d = np.asarray(self.X_df)
        self.y_ser = self.X_ser.copy()
        self.y_df = self.X_df.copy()
        self.y_arr1d = np.asarray(self.y_ser)
        self.y_arr2d = np.asarray(self.y_df)

        self.d = {
            'ser': (self.X_ser, self.y_ser),
            'df': (self.X_df, self.y_df),
            'arr1d': (self.X_arr1d, self.y_arr1d),
            'arr2d': (self.X_arr2d, self.y_arr2d),
        }

    def test_feature_init(self):
        Feature(self.input, self.transformer)

    def _test_robust_transformer(
            self,
            input_types,
            bad_input_checks,
            catches,
            transformer_maker=FragileTransformer):
        fragile_transformer = transformer_maker(bad_input_checks, catches)
        robust_transformer = make_robust_transformer(
            FragileTransformer(bad_input_checks, catches))

        for input_type in input_types:
            X, y = self.d[input_type]
            # fragile transformer raises error
            with self.assertRaises(catches):
                fragile_transformer.fit_transform(X, y)
            # robust transformer does not raise error
            X_robust = robust_transformer.fit_transform(X, y)
            self.assertTrue(
                np.array_equal(
                    asarray2d(X),
                    asarray2d(X_robust)
                )
            )

    def test_robust_transformer_ser(self):
        input_types = ('ser',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_df(self):
        input_types = ('ser', 'df',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_arr(self):
        input_types = ('ser', 'df', 'arr1d')
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
            lambda x: isinstance(x, np.ndarray) and x.ndim == 1,
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_sklearn(self):
        Transformers = (
            SimpleImputer,
            sklearn.preprocessing.StandardScaler,
            sklearn.preprocessing.Binarizer,
            sklearn.preprocessing.PolynomialFeatures,
        )
        # some of these input types are bad for sklearn.
        input_types = ('ser', 'df', 'arr1d')
        for Transformer in Transformers:
            robust_transformer = make_robust_transformer(Transformer())
            for input_type in input_types:
                X, y = self.d[input_type]
                robust_transformer.fit_transform(X, y=y)

    def _test_robust_transformer_pipeline(
            self, input_types, bad_input_checks, catches):
        FragileTransformerPipeline3 = funcy.partial(
            FragileTransformerPipeline, 3)
        return self._test_robust_transformer(
            input_types,
            bad_input_checks,
            catches,
            transformer_maker=FragileTransformerPipeline3)

    def test_robust_transformer_pipeline_ser(self):
        input_types = ('ser',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)

    def test_robust_transformer_pipeline_df(self):
        input_types = ('ser', 'df',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)

    def test_robust_transformer_pipeline_arr(self):
        input_types = ('ser', 'df', 'arr1d')
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
            lambda x: isinstance(x, np.ndarray) and x.ndim == 1,
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)

    def test_feature_as_input_transformer_tuple(self):
        feature = Feature(self.input, self.transformer)
        tup = feature.as_input_transformer_tuple()
        self.assertIsInstance(tup, tuple)
        self.assertEqual(len(tup), 2)

    def test_make_mapper(self):
        feature = Feature(self.input, self.transformer)
        features = [feature]
        mapper = make_mapper(features)
        self.assertIsInstance(mapper, DataFrameMapper)




import numpy as np

from ballet.eng.base import BaseTransformer, GroupedFunctionTransformer

__all__ = ['LagImputer', 'NullFiller', 'NullIndicator']


class LagImputer(GroupedFunctionTransformer):
    def __init__(self, groupby_kwargs=None):
        super().__init__(lambda x: x.fillna(method='ffill'),
                         groupby_kwargs=groupby_kwargs)


class NullFiller(BaseTransformer):
    def __init__(self, isnull=None, replacement=0.0):
        super().__init__()
        if isnull is None:
            self.isnull = np.isnan
        else:
            self.isnull = isnull
        self.replacement = replacement

    def transform(self, X, **transform_kwargs):
        X = X.copy()
        mask = self.isnull(X)
        X[mask] = self.replacement
        return X


class NullIndicator(BaseTransformer):
    def transform(self, X, **tranform_kwargs):
        return np.isnan(X).astype(int)
